@@ -4,9 +4,11 @@
 #include <iostream>
 #include <string>   
 #include <stdexcept>
-
+#ifdef IN_GINAC
+#include "ginac.h"
+#else
 #include <ginac/ginac.h>
-
+#endif
 using namespace std;
 using namespace GiNaC;
 
